import struct


GUIDs = {
    "\x20\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_ConstantForce",
    "\x21\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_RampForce",
    "\x22\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Square",
    "\x23\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Sine",
    "\x24\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Triangle",
    "\x25\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_SawtoothUp",
    "\x26\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_SawtoothDown",
    "\x27\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Spring",
    "\x28\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Damper",
    "\x29\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Inertia",
    "\x2A\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_Friction",
    "\x2B\x1C\x54\x13"+"\x33\x8E"+"\xD0\x11"+"\x9A"+"\xD0"+"\x00"+"\xA0"+"\xC9"+"\xA0"+"\x6E"+"\x35" : "GUID_CustomForce"
        }


def extractString(inp, offset, length):
    return inp[offset:offset+length]
def checkString(inp, key, offset):
    try:
        assert(inp[offset:offset+len(key)] == key)
    except AssertionError:
        print ("Warning, inp[%s:%s+%s] == %s != %s", offset, offset, len(key), key, repr(inp[offset:offset+len(key)]))
def terminateStringOnFirstNullChar(inp):
    end = inp.find("\x00")
    if end > -1:
        return inp[:end]
    else:
        return inp

dataTypes = { \
    "LONG"      : (4, 'l', 0),
    "DWORD"     : (4, 'i', 0)  }

def unpackValue(dataType, inp, offset):
    return struct.unpack('<'+dataTypes[dataType][1], extractString(inp, offset, dataTypes[dataType][0]) + "\x00"*dataTypes[dataType][2])[0]
def packValue(dataType, inp):
    return struct.pack('<'+dataTypes[dataType][1], inp)[:dataTypes[dataType][0]-dataTypes[dataType][2]]

def extractGUID(inp, offset):
    guid = extractString(inp, offset, 16)
    try:
        return GUIDs[guid]
    except KeyError:
        print ("Warning, inp[%s:%s+16]: unknown GUID: %s", offset, offset, repr(guid))

def readFFE(inp):
    results = {}
    print ("Length of file: %s bytes." % len(inp))
    
    """ File Header """
    checkString(inp, "RIFF", 0)
    checkString(inp, "\x00"*4, 0x4)
    checkString(inp, "FORC", 0x8)
    checkString(inp, "\x5C\x77\x7E\x19"+"\xBA\x34\xD3\x11"+"\xAB\xD5\x00\xC0"+"\x4F\x8E\xC6\x27", 0xC)
    """ Start of Effect """
    checkString(inp, "LIST", 0x1C)
    results["effectTotalSize"] = unpackValue("DWORD", inp, 0x20)
    """ Start of Effect String Chunk """
    checkString(inp, "efct", 0x24)
    results["effectName"] = terminateStringOnFirstNullChar(extractString(inp, 0x28, 0x68-0x28)) # Max length = 63 + terminatorChar
    # Place left here may be filled with garbage (after null char).
    """ Start of Effect Struct """
    results["effectStructSize"] = unpackValue("DWORD", inp, 0x68) # Including this DWORD (= len 4)
    results["dwFlags"] = unpackValue("DWORD", inp, 0x6C)
    results["duration"] = unpackValue("DWORD", inp, 0x70)
    results["samplePeriod"] = unpackValue("DWORD", inp, 0x74)
    results["gain"] = unpackValue("DWORD", inp, 0x78)
    results["triggerButton"] = unpackValue("DWORD", inp, 0x7C)
    results["triggerRepeatInterval"] = unpackValue("DWORD", inp, 0x80)
    results["cAxes"] = unpackValue("DWORD", inp, 0x84)
    results["rgdwAxes?"] = unpackValue("LONG", inp, 0x88); checkString(inp, "\x20\x00\x00\x00", 0x88) # Seems to be always equal to "\x20" = 32
    results["rglDirection?"] = unpackValue("LONG", inp, 0x8C); checkString(inp, "\x40\x00\x00\x00", 0x8C) # Seems to be always equal to "\x40" = 64
    results["lpEnvelope?"] = unpackValue("LONG", inp, 0x90); checkString(inp, "\x80\x00\x00\x00", 0x90) # Seems to be always equal to "\x80" = 128
    results["cbTypeSpecificParams = typeSpecificStructSize"] = unpackValue("LONG", inp, 0x94)
    results["lpvTypeSpecificParams?"] = unpackValue("LONG", inp, 0x98); checkString(inp, "\x00\x01\x00\x00", 0x98) # Seems to be always equal to "\x00\x01" = 256
    results["startDelay"] = unpackValue("DWORD", inp, 0x9C)
    """ Start of GUID """
    results["GUID"] = extractGUID(inp, 0xA0)
    checkString(inp, "\x01\x00\x00\x00", 0xB0)
    """ Start of Axes array """    #           <-- Until here, absolute size
    checkString(inp, "data", 0xB4)
    results["axesArraySize"] = unpackValue("DWORD", inp, 0xB8)
    results["axes[0]"] = unpackValue("DWORD", inp, 0xBC) # Axis 0
    """ Start of Direction array """
    checkString(inp, "data", 0xC0)
    results["directionArraySize"] = unpackValue("DWORD", inp, 0xC4)
    results["direction[0]"] = unpackValue("LONG", inp, 0xC8) # Axis 0
    """ Start of Envelope Struct """
    checkString(inp, "data", 0xCC)
    results["envelopeStructSize = envelope.dwSize"] = unpackValue("DWORD", inp, 0xD0); assert(inp[0xD0:0xD0+4] == inp[0xD4:0xD4+4])
    results["envelope.dwSize"] = unpackValue("DWORD", inp, 0xD4) # Including this DWORD (= len 4)
    results["attackLevel"] = unpackValue("DWORD", inp, 0xD8)
    results["attackTime"] = unpackValue("DWORD", inp, 0xDC)
    results["fadeLevel"] = unpackValue("DWORD", inp, 0xE0)
    results["fadeTime"] = unpackValue("DWORD", inp, 0xE4)
    """ Start of Type Specific Struct """
    checkString(inp, "data", 0xE8)
    results["typeSpecificStructSize"] = unpackValue("DWORD", inp, 0xEC); assert(inp[0x94:0x94+4] == inp[0xEC:0xEC+4])
    results["magnitude"] = unpackValue("LONG", inp, 0xF0)
    
    return results

print (readFFE(open("./TestData/FEditTests/FEdit4.ffe", 'rb').read()))
